/* 
   sitecopy, for managing remote web sites.
   Copyright (C) 1998-99, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* This is the console front end for sitecopy.
 * TODO: Some of the configuration stuff should be moved out of here. */

#include "config.h"

#include <sys/types.h>

#include <sys/param.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>

#ifndef HAVE_GETOPT
#include "getopt.h"
#endif

#include "sites.h"
#include "rcfile.h"
#include "common.h"

/* The defined sites from sites.c */
extern struct site_t *all_sites;

/* The maximum number of sites which can be specified on the command
 * line */

#define MAXSITES 20

enum action_t {
    action_none,
    action_list,
    action_synch,
    action_fetch,
    action_update,
    action_catchup,
    action_init,
    action_view
} action;

extern int rcfile_linenum;
extern char *rcfile_line;

/* Misc global vars */
extern char *copypath; /* path to info file storage directory */
extern char *progname; /* argv[0] */
extern char *rcfile; /* filename of rcfile */

char *sitenames[MAXSITES]; /* the sites specified on the command line */
int numsites; /* the number of sites specified */
struct site_t *current_site; /* this is used to save the state if we
			      * get signalled mid-update */
bool needsite; /* Is a site name required for the given action? */

/* User-specified options */
extern bool fe_prompting; /* Do they want confirmations for update mode? */
bool allsites; /* Do they want all sites to be operated on? */
bool listflat; /* Do they want the 'flat' list style */
int quiet; /* How quiet do they want us to be? */

/* Functions prototypes */

void init( int, char ** );
int main( int, char ** );
void usage( );
void site_flatlist_items( struct site_t *, const enum file_diff,
			  const char * );
void site_normlist( struct site_t * );
void site_flatlist( struct site_t * );

void sighandler( int sig ) {
#if 0
    printf( "\n\n%s: Caught signal: %s (%d)\n", progname, 
	    strsignal( sig ), sig );
#else
    printf( "\n\n%s: Caught signal %d\n", progname, sig );
#endif
    if( action == action_update && current_site!=NULL ) {
	printf( "%s: In site update, saving current progress... ", progname );
	fflush( stdout );
	site_writefiles( current_site );
	printf( "done.\n" );
    }
    printf( "%s: Terminating.\n", progname );
    exit( 15 );
}

/* Produce the normal listing output for the given site.
 */
void site_normlist( struct site_t *the_site ) {
    struct site_file_t *current;
    int count;

    if( the_site->numnew > 0 ) {
	printf( "* These items have been added since the last update:\n" );
	count = 0;
	for( current = the_site->files; current!=NULL; current=current->next) {
	    if( current->diff == file_new ) {
		if( count++ ) printf( ", " );
		if( current->dir ) printf( "dir:" );
		    printf( "%s", current->rel_local);
	    }
	}
	putchar( '\n' );
    }
    if( the_site->numchanged > 0 ) {
	printf( "* These items have been changed since the last update:\n" );
	count = 0;
	for( current = the_site->files; current!=NULL; current=current->next) {
	    if( current->diff == file_changed ) {
		if( count++ ) printf( ", " );
		printf( "%s", current->rel_local);
	    }
	}
	putchar( '\n' );
    }
    if( the_site->numdeleted > 0 ) {
	printf( "* These items have been deleted since the last update:\n" );
	count = 0;
	for (current = the_site->files; current!=NULL; current=current->next) {
	    if( current->diff == file_deleted ) {
		if( count++ ) printf( ", " );
		if( current->dir ) printf( "dir:" );
		printf( "%s", current->rel_local );
	    }
	}
	if( the_site->nodelete ) {
	    printf( "-- These files will be left on the server --\n" );
	}
	putchar( '\n' );
    }
    if( the_site->nummoved > 0 ) {
	printf( "* These items have been moved since the last update:\n" );
	count = 0;
	for (current = the_site->files; current!=NULL; current=current->next) {
	    if( current->diff == file_moved ) {
		if( count++ ) printf( ", " );
		printf( "%s->/%s", current->old->rel_local, 
			current->directory  );
	    }
	}
	putchar( '\n' );
    }
}


/* Produces a section of the flat listing output, of all the items
 * with the given diff type in the given site, using the given section
 * name. */
void site_flatlist_items( struct site_t *the_site, const enum file_diff diff,
			  const char *name ) {
    struct site_file_t *current;
    printf( "sectstart|%s", name );
    putchar( '\n' );
    for( current = the_site->files; current!=NULL; current=current->next) {
	if( current->diff == diff ) {
	    printf( "item|%s%s", current->rel_remote,
		    current->dir?"/":"" );
	    if( current->old !=NULL ) {
		printf( "|%s\n", current->old->rel_remote );
	    } else {
		putchar( '\n' );
	    }	    
	}
    }
    printf( "sectend|%s\n", name );
}

/* Produce the flat listing output for the given site */
void site_flatlist( struct site_t *the_site ) {
    printf( "sitestart|%s", the_site->name );
    if( the_site->url )	printf( "|%s", the_site->url );
    putchar( '\n' );
    if( the_site->numnew > 0 )
	site_flatlist_items( the_site, file_new, "added" );
    if( the_site->numchanged > 0 )
	site_flatlist_items( the_site, file_changed, "changed" );
    if( the_site->numdeleted > 0 )
	site_flatlist_items( the_site, file_deleted, "deleted" );
    if( the_site->nummoved > 0 )
	site_flatlist_items( the_site, file_moved, "moved" );
    printf( "siteend|%s\n", the_site->has_changed?"changed":"unchanged" );
}

void site_dumpsites( ) {
    struct site_t *current;
    for( current=all_sites; current!=NULL; current=current->next ) {
	/* FIXME: Possibly, make sure we have the actual port number first
	 * so we don't have to mess around printing out the port */
	printf( "Site: %s\n"  "\tServer: %s", current->name, current->server );
	if( current->port == 0 ) {
	    printf( "  Port: %s\n", current->driver->service_name );
	} else {
	    printf( "  Port: %d\n", current->port );
	}
	printf( "\tProtocol: %s    Username: %s\n",
		current->driver->protocol_name, current->username );
	if( ! current->ftp_pasv_mode )
	    printf( "\tPassive mode FTP will not be used.\n" );
	printf( "\tRemote directory: %s%s\n" 
		"\tLocal directory: %s\n",
		DIRPREFIX(current->remote_root), current->remote_root, 
		current->local_root );
	printf( "\tPermissions: %s     Symlinks: %s\n",
		PERMSMODE( current ), SYMLINKSMODE( current ) );
	if( current->nodelete ) 
	    printf( "\tRemote files will not be deleted.\n" );
	if( current->checkmoved ) 
	    printf( "\tFiles will be moved remotely if moved locally.\n" );
    }
}

void init( int argc, char *argv[] ) {
    int optc;
    bool firstlist = false;
    extern char *optarg;
    extern int optind;
#ifdef DEBUGGING
    extern int debug_mask;

    debug_mask = 0;
#endif
    /* Defaults */
    allsites = fe_prompting = false;
    action = action_list;
    fe_prompting = false;
    progname = basename( argv[0] );
    needsite = false;

    /* Read the cmdline args */
    while( (optc = getopt( argc, argv, 
#ifdef DEBUGGING
"d:"
#endif
"acifhlp:r:suvy" ) ) != -1) {
	switch( optc ) {
	case 'l': 
	    action = action_list;
	    if( firstlist == false ) {
		firstlist = true;
	    } else {
		listflat = true;
	    }
	    break;
	case 's': action = action_synch; break;
	case 'f': action = action_fetch; break;
	case 'i': action = action_init; break;
	case 'c': action = action_catchup; break;
	case 'u': action = action_update; break;
	case 'v': 
	    action = action_view;
	    allsites = true; /* so they get checked */
	    break;
	case 'q': quiet++; break;
	case 'r': 
	    if( strlen(optarg) != 0 ) {
		rcfile = strdup( optarg );
	    } else {
		usage( );
		exit(-1);
	    }
	    break;
	case 'p':
	    if( strlen(optarg) != 0 ) {
		if( optarg[strlen(optarg)] != '/' ) {
		    copypath = malloc( strlen(optarg) + 2 );
		    strcpy( copypath, optarg );
		    strcat( copypath, "/" );
		} else {
		    copypath = strdup( optarg );
		}
	    } else {
		usage( );
		exit(-1);
	    }
	    break;
#ifdef DEBUGGING
	case 'd': /* set debugging level */
	    debug_mask = atoi(optarg);
	    break;
#endif

	case 'y':
	    fe_prompting = true;
	    printf( "Prompting on.\n" );
	    break;
	case 'a':
	    allsites = true;
	    break;
	case 'h': usage( ); exit(-1);
	case '?': usage( ); exit(-1);
	}
    }

    /* Determine whether the specified action needs a site name
     * to operate on */
    needsite = ((action==action_list) || (action==action_update) || \
		(action==action_catchup) || (action==action_synch) || \
		(action==action_fetch) || (action==action_init));
    
    /* Get those site names off the end of the cmdline */
    for( numsites = 0 ; optind < argc; optind++ ) {
	sitenames[numsites++] = argv[optind];
	if( numsites == MAXSITES ) {
	    printf( "%s: Warning: Only %d sites can be specified on the command line!\nExtra entries are being skipped.\n", progname, MAXSITES );
	    break;
	}
    }

}

/* Prompts for confirmation - returns */
bool fe_ynprompt( ) {
    char in[BUFSIZ];
    fgets( in, BUFSIZ, stdin );
    return (*in == 'y'); /* only return true if they put a 'y' first */
}

bool fe_can_update( struct site_file_t *file ) {
    if( file->dir ) {
	if( file->diff == file_new ) {
	    printf( "Create %s/", file->rel_local );
	} else {
	    printf( "Delete %s/", file->rel_local );
	}
    } else {
	switch( file->diff ) {
	case file_changed:
	case file_new: printf( "Upload %s", file->rel_local ); break;
	case file_deleted: printf( "Delete %s", file->rel_local ); break;
	case file_moved: printf( "Move %s->/%s", file->old->rel_local,
				 file->directory ); break;
	default: printf( "You've really broken it now.\n" ); break;
	}
    }
    printf( "? (y/n) " );
    return fe_ynprompt( );
}

/* Called when the given files is about to be updated */
void fe_updating( struct site_file_t *file ) {
    switch( quiet ) {
    case 0:
	if( file->dir ) {
	    if( file->diff == file_new ) {
		printf( "Creating %s/: ", file->rel_local );
	    } else {
		printf( "Deleting %s/: ", file->rel_local );
	    }
	} else {
	    switch( file->diff ) {
	    case file_changed:
	    case file_new: printf( "Uploading %s: [", file->rel_local ); break;
	    case file_deleted: printf( "Deleting %s: ", file->rel_local ); break;
	    case file_moved: printf( "Moving %s->/%s: ", file->old->rel_local,
				     file->directory ); break;
	    default: printf( "You've really broken it now.\n" ); break;
	    }
	}
	fflush( stdout );
	break;
    case 1:
	printf( "%s\n", file->rel_local );
	break;
    default:
	break;
    }
}

void fe_updated( const struct site_file_t *file, const bool success,
		 const char *error ) {
    if( quiet > 0 )
	return;
    if( file->dir || (file->diff!=file_changed && file->diff!=file_new) ) {
	if( success ) {
	    printf( "done.\n" );
	} else {
	    printf( "failed:\n%s\n", error );
	}
    } else {
	if( success ) {
	    printf( "] done.\n" );
	} else {
	    printf( "] failed:\n%s\n", error );
	}
    }
}

void fe_synching( struct site_file_t *file ) {
    if( quiet ) {
	printf( "%s\n", file->rel_local );
    } else {
	if( file->dir ) {
	    if( file->diff != file_new ) {
		printf( "Creating %s/: ", file->rel_local );
	    } else {
		printf( "Deleting %s/: ", file->rel_local );
	    }
	} else {
	    switch( file->diff ) {
	    case file_changed:
	    case file_deleted: 
		printf( "Downloading %s: [", file->rel_local ); 
		break;
	    case file_new: 
		printf( "Deleting %s: ", file->rel_local ); 
		break;
	    case file_moved: 
		printf( "Moving %s->/%s: ", file->rel_local,
			file->old->directory ); 
		break;
	    default: 
		break;
	    }
	}
	fflush( stdout );
    }
}

void fe_synched( const struct site_file_t *file, const bool success,
		 const char *error ) {
    if( quiet )
	return;
    if( file->dir || (file->diff!=file_changed && file->diff!=file_deleted) ) {
	if( success ) {
	    printf( "done.\n" );
	} else {
	    printf( "failed:\n%s\n", error );
	}
    } else {
	if( success ) {
	    printf( "] done.\n" );
	} else {
	    printf( "] failed:\n%s\n", error );
	}
    }
}

/* Hidden tricks:
 * uncomment the #define below for a progress spinner rather than the
 * dots.
 */

/* #define PROGRESS_SPINNER */

void fe_transfer_progress( int num, int total ) {
#ifndef PROGRESS_SPINNER
    putchar( '.' );
    fflush( stdout );
#else
    static int ty;
    if( ty > 0 ) {
	putchar( '\b' );
    } else {
	ty = 1;
    }
    if( ty < 2 ) {
	putchar( '|' );
    } else {
	putchar( '-' );
    }
    ty = (ty % 2) + 1;
    fflush( stdout );
#endif
}

/* Not really implemented.
void fe_warning( int num ) {
    printf( "Warning: %d\n", num );
}
*/

void init_sites( ) {
    int ret;
    struct site_t *current;
    
    /* Read the rcfile */
    ret = rcfile_read( &all_sites );
    if( ret == RC_OPENFILE ) {
	printf( "%s: Error: Could not open rcfile: %s.\n",
		progname, rcfile );
	exit( -1 );
    } else if( ret == RC_CORRUPT ) {
	printf( "%s: rcfile corrupt at line %d:\n%s",
		progname, rcfile_linenum, rcfile_line );
	exit( -1 );
    }
    /* Now go through the list and verify all the sites are valid. */
    for( current = all_sites; current!=NULL; current=current->next ) {
	ret = rc_verifysite( current );
	if( ret == 0 ) continue;
	switch( ret ) {
	case SITE_ACCESSLOCALDIR:
	    printf( "%s: Could not read directory for `%s':\n\t%s\n", 
		    progname, current->name, current->local_root );
	    break;
	case SITE_NOSERVER:
	    printf( "%s: Server not specified in site `%s'.\n", 
		    progname, current->name );
	    break;
	case SITE_NOUSERNAME:
	    printf( "%s: Username not specified in site `%s'.\n", 
		    progname, current->name );
	    break;
	case SITE_NOPASSWORD:
	    printf( "%s: Password not specified in site `%s'.\n", 
		    progname, current->name );
	    break;
	case SITE_NOREMOTEDIR:
	    printf( "%s: Remote directory not specified in site `%s'.\n", 
		    progname, current->name );
	    break;
	case SITE_NOLOCALDIR:
	    printf( "%s: Local directory not specified in site `%s'.\n", 
		    progname, current->name );
	    break;
	case SITE_INVALIDPORT:
	    printf( "%s: Invalid port used in site `%s'.\n",
		    progname, current->name );
	    break;
	case SITE_NOMAINTAIN:
	    printf( "%s: %s cannot maintain symbolic links (site `%s').\n",
		    progname, current->driver->protocol_name, current->name );
	    break;
	default:
	    printf( "%s: Unhandled error in site %s, please contact the maintainer.\n", progname, current->name );
	    break;
	    
	}
	exit( -1 );
    }
}

int main( int argc, char *argv[] ) {
    int ret, num, numgoodsites;
    struct site_t *current;

    /* Initialize up */
    ret = 0;

    init( argc, argv );

    if( init_env( ) == 1 ) {
	printf( "%s: Error: Environment variable HOME not set.\n",
		progname );
	exit( -1 );
    }

    ret = init_paths( );

    if( ret != 0 ) {
	switch( ret ) {
	case RC_OPENFILE:
	    printf( "%s: Error: Could not open rcfile.\n", progname );
	    break;
	case RC_PERMS:
	    printf( "%s: Error: rcfile permissions allow other users to read your rcfile.\n", progname );
	    break;
	case RC_DIROPEN:
	    printf( "%s: Error: Could not open info directory.\n", progname );
	    break;
	case RC_DIRPERMS:
	    printf( "%s: Error: info directory permissions incorrect.\n", progname );
	    break;
	case RC_NETRCPERMS:
	    printf( "%s: Error: ~/.netrc permissions incorrect.\n", progname );
	    break;
	default:
	    break;
	}
	exit(-1);
    }

    if( init_netrc( ) == 1 ) {
	printf( "%s: Error: Could not parse ~/.netrc.\n", progname );
	exit( -1 );
    }

    ret = 0;

    init_sites( );

    /* Handle signals, so if we get killed in the middle of a
     * site update, the state gets saved properly. */
    signal( SIGPIPE, sighandler );
    signal( SIGTERM, sighandler );
    signal( SIGQUIT, sighandler );
    signal( SIGINT, sighandler );
    signal( SIGABRT, sighandler );

    if( allsites ) {
	
	for( current = all_sites; current!=NULL; current=current->next ) {
	    if( site_readfiles( current ) ) {
		printf( "%s: Error: Could not read site info file for `%s'.\n",
			progname, current->name );
		numgoodsites--;
	    } else {
		/* Got a valid site! */
		current->use_this = true;
	    }
	}

    } else if( needsite ) {
	/* Find the site they want */
        
	numgoodsites = numsites;

	for( num=0; num < numsites; num++ ) {
	    current = site_find( sitenames[num] );
	    if( current == NULL ) { 
		printf( "%s: Error: No site called `%s' found - skipping.\n", 
			progname, sitenames[num] );
		numgoodsites--;
	    } else if( site_readfiles( current ) ) {
		printf( "%s: Error: Could not read site info file for `%s'.\n",
			progname, sitenames[num] );
		numgoodsites--;
	    } else {
		/* Got a valid site! */
		current->use_this = true;
	    }
	}
	
	if( numgoodsites==0 ) {
	    printf( "%s: No sites specified.\n", progname );
	    exit(-1);
	}
		
    }

    /* Handle the dump sites request */
    if( ! needsite ) {
	switch( action ) {
	case action_view:
	    site_dumpsites( );
	    ret = 0;
	    break;
	default:
	    printf( "Unimplemented.\n" );
	    ret = -1;
	}
	return ret;
    }

    for( current=all_sites; current!=NULL; current=current->next ) {
	
	if( !current->use_this ) 
	    continue;
	
	if( !listflat ) {
	    printf( "%s: Site `%s' (%s://%s%s%s) in %s\n", 
		progname, current->name, 
		current->driver->service_name, current->server, 
		DIRPREFIX(current->remote_root), 
		current->remote_root, current->local_root );
	}

	switch( action ) {
	case action_update:
	    if( current->has_changed ) {
		current_site = current;
		ret = site_update( current );
		current_site = NULL; /* hope we don't get signalled here */
		site_writefiles( current );
	    } else {
		printf( "%s: Nothing to do - no changes found.\n", progname );
		ret = 0;
	    }
	    break;
	case action_list:
	    if( listflat ) {
		site_flatlist( current );
	    } else {
		site_normlist( current );
	    }
	    ret = (current->has_changed)?1:0;
	    if( current->has_changed ) {
		if( !listflat ) {
		    printf( "%s: The remote site needs updating.\n", progname);
		}
		ret = 1;
	    } else {
		if( !listflat ) {
		    printf( "%s: The remote site does not need updating.\n", progname );
		}
		ret = 0;
	    }
	    break;
	case action_init:
	    printf( "Initializing site... " );
	    site_initialize( current );
	    site_writefiles( current );
	    printf( "done.\n" "All the files and directories are marked as NOT updated remotely.\n" );
	    break;
	case action_catchup:
	    printf( "Catching up site... " );
	    site_catchup( current );
	    site_writefiles( current );
	    printf( "done.\n" "All the files and and directories are marked as updated remotely.\n" );
	    break;
	case action_fetch:
	    printf( "Fetching remote site listing... " );
	    ret = site_fetch( current );
	    switch( ret ) {
	    case SITE_UNIMPL: 
		printf( "\nUnimplemented for %s, sorry.\n",
			current->driver->protocol_name );
		ret = -1;
		break;
	    default:
		site_writefiles( current );
		printf( "done.\n" );
	    } 
	    break;
	case action_synch:
	    if( current->has_changed ) {
		printf( "Resynchronizing site...\n" );
		site_synch( current );
		printf( "Resynchronize finished.\n" );
	    } else {
		printf( "%s: Nothing to do - no changes found.\n", progname );
	    }
	    break;
	default:
	    printf( "Unimplemented - sorry.\n");
	    ret = -1;
	    break;
	}
    }
    
    return ret;
}

void usage( ) {
    fprintf( stderr,
"Usage: %s [options] [operation mode] [sitename...]\n"
"Options: "
#ifdef DEBUGGING
"[-d x] "
#endif
"[-y] [-r fn] [-p dir]\n\n"
#ifdef DEBUGGING
"\t-d\tSet debugging level - see documentation\n"
#endif
"\t-r\tSpecify alternate run control file\n"
"\t-p\tSpecify alternate site storage directory\n"
"\t-y\tRequest confirmation before making each update\n"
"\t-v\tList the site definitions\n"
"\t-a\tPerform the operation on ALL defined sites.\n"
"\nOperation modes:\n"
"\t-h\tDisplay this help message\n"
"\t-l\tList files which have been added, changed, moved and deleted\n"
"\t-i\tInitialize - mark all files and directories as not updated\n"
"\t-f\tFetch - find out what files are on the remote site\n"
"\t-s\tSynchronize - update the local site from the remote site ***\n"
"\t-c\tCatch up - mark all files and directories as updated\n"
"\t-u\tUpdate the remote site\n"
"\n\t(***) This operation will overwrite local files - USE WITH CARE\n"
"\nI am " PACKAGE " version %s, console user interface"
#ifdef DEBUGGING
", debugging.\n"
#else
".\n"
#endif
"\tProtocols supported: FTP"
#ifdef USE_DAV
", WebDAV"
#endif
"\n\n", progname, VERSION );

/* \t-a\tPerform the given action on ALL sites defined\n\
   \t-r\tResyncronize the remote site list\n\
*/

}
/* "Warning: directory search depth exceeded. Increase MAXDIRS in sites.c.\n\
   Directory %s is being skipped.\n" */
