/* 
   sitecopy, for managing remote web sites.
   Copyright (C) 1999, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* TODO:
 * - Make the 'remote' and 'local' file state more general,
 *   allow for permissions being being different etc etc.
 *   Necessary for accurate synchronization.
 * - The 'driver' and 'driver_options' thing is all a bit crap.
 *   I'm not sure how to do this better, ideas welcome.
 */

#ifndef SITES_H
#define SITES_H

#include <time.h>

#include "common.h"
#include "protocol.h"

#define MAXEXCLUDES 20

/* Macro to remove items from a doubly linked list...
 * give it the item to delete and the list head pointer */
#define LL_DELETE(item,head)			\
if( item->prev ) {				\
    item->prev->next = item->next;		\
} else {					\
    head = item->next;				\
}						\
if( item->next ) {				\
    item->next->prev = item->prev;		\
}

/* #define DIRPREFIX(item) (item[0]!='/')?"~/":"" */
#define DIRPREFIX(item) (item[0]!='/')?"/":""

#define PERMSMODE(site) \
(site->perms==sitep_ignore)?"Ignored":(site->perms==sitep_exec)?"Maintained for executables":"Always maintained"
#define SYMLINKSMODE(site) \
(site->symlinks==sitesym_ignore?"Ignored":"Followed")

/* Unimplemented function of protocol driver */
#define SITE_UNIMPL 931

struct site_file_t;
struct site_t;

enum file_diff {
    file_unchanged, /* Remote file is same as local file */
    file_changed, /* File has changed locally, and should be uploaded */
    file_new, /* File is new locally, and should be uploaded */
    file_deleted,  /* File deleted locally, and should be deleted remotely */
    file_moved /* File has been moved locally, should be moved remotely */
};

/* Valid file permissions mirroring values */
enum site_perm_modes {
    sitep_ignore, /* Ignore file permissions */
    sitep_exec, /* Maintain execute permissions */
    sitep_all /* Maintain all permissions */
};

/* Valid symlink handling modes */
enum site_symlink_modes {
    sitesym_ignore,
    sitesym_follow,
    sitesym_maintain
};

/* Protocol modes */

enum site_protocol_modes {
    siteproto_ftp,
    siteproto_http
};

/* Linked list for holding filename mappings.
 * Mappings are held for the 'name' part of filenames only (see below).
 */
struct name_mapping_t {
    char *local;
    char *remote;
    struct name_mapping_t *next;
};

/* A filename is made up of three parts:
   The site root path, the directory, and the name.
   eg, a remote file:

   /html/mainsite/subdir/another/file.html
   [--site root--][--directory--][-name--]

   site root = "html/mainsite/"
   directory = "subdir/another/"
   name = "file.html"

   The directory is the same remotely as locally, eg, a remote file cannot
   be mapped to a local file in a different directory. A directory MUST
   NOT include a leading slash but MUST include a trailing slash, except in
   the special case of a file in the site root directory, in which case the
   directory part will be zero-length.

   Names may be different remotely and locally.

   Within a given site, the site roots are the same for all files locally
   and remotely. The site roots may be 0-length, eg. for FTP sites where
   the home (login) directory is the site root directory.

   Note that a directory is also classed as any other 'file'.

   This gives us three distinct properties for each file:
     - The directory
     - The local name
     - The remote name
   
   For convenience, the site_file_t structure contains four other properties
   which are *directly* derived from the three properties, and the site root:
     - The 'full' local filename
     - The 'full' remote filename
     - The 'relative' local filename
     - The 'relative' remote filename
   
   The 'full' filenames are:
      site root + directory + (local or remote) name
   The 'relative' filenames are:
      "/" + directory + (local or remote) name
  
   The 'relative' filenames for displaying to users.
   The 'full' filenames are for communicating with the server and use
   with the local file system.

   The site_assignnames() function fills the full_* and rel_* members.
   
   Final note: a file is excluded if it's LOCAL name matches an
   exclude pattern. This is relevant to fetch mode, where the remote
   name might not match a pattern but it may be mapped to a local name
   which does. Pretty damn obscure, but hey.

*/

struct site_file_t {
    enum file_diff diff;
    bool dir; /* This is a directory */
    bool link; /* This is a link */

    char *directory; /* the directory, relative to site root */
    char *remotename; /* the remote name */
    char *localname; /* the local name */

    /* The following two are computed from the above three for
     * simplicity, and are the FULL remote and local file names */
    char *full_remote;
    char *full_local; 
    /* These are for convenience, and is directory cat'ed with localname */
    char *rel_local;
    char *rel_remote;

    struct site_file_t *old; /* the item before it was moved */

    time_t localtime; /* the local last mod time of the file */
    time_t remotetime; /* the remote last mod time of the file */
    off_t localsize; /* the local file size */
    off_t remotesize; /* the remote file size */
    mode_t mode; /* the permissions */
    bool updated; /* whether the remote file has been updated according
		   * to the diff setting */

    char *remotelink; /* the remote target of the link */    
    char *locallink; /* the local target of the link */

    /* Linked list nodes */
    struct site_file_t *next;
    struct site_file_t *prev;

};

/* This represents a site */
struct site_t {

    char *name; /* symbolic name for site */
    char *url; /* URL for site - used by -ll mode */

    char *server; /* server name */
    int port; /* server port */
    char *username; /* username to log in to server */
    char *password; /* password for ditto */

    enum site_protocol_modes protocol;
    struct proto_driver *driver; /* the protocol driver routines */

    /* ugh - protocol specific options here */
    bool ftp_pasv_mode;
    
/*  UNUSED currently:
    bool http_expect_continue;
    bool http_persistent; */
    
    char *remote_root; /* root directory of site on server */
    char *local_root; /* root directory of site locally */
    char *infofile;  /* local storage file in ~/.sitecopy/  */

    /* Options for the site */
    enum site_perm_modes perms; /* permissions maintenance mode */
    enum site_symlink_modes symlinks; /* symlink handline mode */
    bool nodelete;
    bool checkmoved;
    
    /* array of shell regex's of files to exclude */
    char *excludes[MAXEXCLUDES]; 
    int numexcludes; /* number of files in above array */

    /* array of renames */
    struct name_mapping_t *renames;

    struct site_file_t *files; /* list of files */
    struct site_file_t *files_tail; /* end of the list */

    /* Some useful counts for the files */
    int numnew; /* number of new files */
    int numchanged; /* number of changed files */
    int numdeleted; /* number of deleted files */
    int nummoved; /* number of moved files */
    int numunchanged; /* number of unchanged files */
    
    size_t totalnew; /* total file size of new files */
    size_t totalchanged; /* total file size of changed files */

    bool has_changed; /* whether the site has actually changed */
    bool use_this; /* whether the site is being operated on */

    struct site_t *next;

};

/* Internal use - assigns the local and remote filenames to the site file */
void site_assignnames( struct site_file_t *the_file, struct site_t *the_site );

/* This updates the files information for the given site - both the
 * local and remote ones. Returns 0 on success, or non-zero on failure */
int site_readfiles( struct site_t * );

/* This marks remote files as updated locally */
int site_catchup( struct site_t * );

/* This writes the remote files list back to the file */
int site_writefiles( struct site_t * );

/* Initialize the site - pretend there are NO files held remotely */
int site_initialize( struct site_t * );

/* Updates the remote files list of the given site.
 * Tricky, this - have to do a load of LIST's, and try and
 * parse the response sensibly */
int site_update( struct site_t * );

/* Finds a site with the given name, and returns a pointer to it.
 * If no site of given name is found, returns NULL
 */
struct site_t *site_find( char *sitename );

void site_checkmoved( struct site_t *any_site );

/* Syncronizes the remote site with the local copy.
 * This is the core functionality of sitecopy. */
int site_synch( struct site_t *the_site );

/* Updates the files listing from the remote site. */
int site_fetch( struct site_t *the_site );

/* Destroys all the files... use before doing a second
 * site_readfiles on a site. */
void site_destroy( struct site_t *the_site );

#endif /* SITES_H */
